; pwm-sound.inc.asm
; Erzeugung von Siuns- und Oberwellen ber einen PWM-Ausgang oder (falls angegeben) einen Pin, der durch
; .equ PWMSOUND_PORT = <PORTB|PORTC|PORTD>
; .equ PWMSOUND_PIN = <0..7>
; konfiguriert wird.
;
; Die Routinen arbeiten mit den beiden 8-Bit-Timern 0 und 2. Timer 0 ruft 31250 mal/sek eine ISR auf,
; die das nchste Sample berechnet, aus dem Timer2 einen PWM-Wert erzeugt.
; Abhngig davon, ob gerade ein Ton gespielt wird, kann, falls angegeben ein Pin (z.B. zum aus/einschalten 
; des Verstrkers) ein- oder ausgeschaltet werden.
; Dieser wird durch
; .equ PWMSOUND_ONOFF_PORT = <PORTB|PORTC|PORTD>
; .equ PWMSOUND_ONOFF_PIN = <0..7>
; .equ PWMSOUND_ONOFF_ON = 1|0 ; Wert, auf den der Pin gesetzt werden muss, wenn der Ton gespielt wird
; definiert.

.ifdef PWMSOUND_TIMER
 .if (PWMSOUND_TIMER == 1) || (PWMSOUND_TIMER == 1)
.else
 .error "PWMSOUND_TIMER muss definiert sein!"
.endif

.ifdef PWMSOUND_PORT
 .if ((PWMSOUND_PORT == PORTB) || (PWMSOUND_PORT == PORTC) || (PWMSOUND_PORT == PORTD))
  .ifdef PWMSOUND_PIN
    .if ((PWMSOUND_PIN >= 0) || (PWMSOUND_PIN <= 7))
     .equ PWMSOUND_PORTAUSGABE = 1
    .else
     .error "PWMSOUND_PIN muss einen Wert zwischen 0 und 7 haben!"
   .endif
  .else
   .error "Wenn PWMSOUND_PORT angegeben ist, muss auch PWMSOUND_PIN angegeben werden."
  .endif
 .else
  .error "PWMSOUND_PORT muss ein gltiger Port (z.B. PORTD) sein.
 .endif
.else
 .equ PWMSOUND_PORTAUSGABE = 0
.endif

.ifdef PWMSOUND_ONOFF_PORT
 .if ((PWMSOUND_ONOFF_PORT == PORTB) || (PWMSOUND_ONOFF_PORT == PORTC) || (PWMSOUND_ONOFF_PORT == PORTD))
  .ifdef PWMSOUND_ONOFF_PIN
    .if ((PWMSOUND_ONOFF_PIN >= 0) || (PWMSOUND_ONOFF_PIN <= 7))
	 .ifdef PWMSOUND_ONOFF_ON
	  .if ((PWMSOUND_ONOFF_ON == 0) || (PWMSOUND_ONOFF_ON == 1))
       .equ PWMSOUND_ONOFF_SCHALTEN = 1
      .else
       .error "PWMSOUND_ONOFF_PIN muss einen Wert zwischen 0 und 7 haben!"
      .endif
     .else
	  .error "Wenn PWMSOUND_ONOFF_PORT angegeben ist, muss auch PWMSOUND_ONOFF_ON angegeben werden."
	 .endif
    .else
     .error "PWMSOUND_ONOFF_PIN muss einen Wert zwischen 0 und 7 haben!"
   .endif
  .else
   .error "Wenn PWMSOUND_ONOFF_PORT angegeben ist, muss auch PWMSOUND_ONOFF_PIN angegeben werden."
  .endif
 .else
  .error "PWMSOUND_PORT muss ein gltiger Port (z.B. PORTD) sein.
 .endif
.else
 .equ PWMSOUND_ONOFF_SCHALTEN = 0
.endif

.if (PWMSOUND_PORTAUSGABE)
 .equ DDR_PWMSOUND = PWMSOUND_PORT - 1
 .equ PWMSOUND_COM2 = 0 ; Normal port operation, OC2 disconnected.
 ;.equ PWMSOUND_WGM2 = 0 ; Normal
 .equ PWMSOUND_WGM2 = 3 ; fast PWM
 .equ PWMSOUND_TIMSK = 	((1<<TOIE0)|(1<<TOIE2)|(1<<OCIE2))
.else
 .equ PWMSOUND_COM2 = 2 ; Clear OC2 on Compare Match, set OC2 at TOP
 .equ PWMSOUND_WGM2 = 3 ; fast PWM
 .equ PWMSOUND_TIMSK = 	((1<<TOIE0)|(1<<TOIE2))
.endif
.equ PWMSOUND_PRESCALING2 = 1 ; no Prescaling
.equ PWMSOUND_WGM2_BITMUSTER = ((PWMSOUND_WGM2 & 2 >> 1) << WGM21) | ((PWMSOUND_WGM2 & 1) << WGM20)

.if (PWMSOUND_ONOFF_SCHALTEN)
 .equ DDR_PWMSOUND_ONOFF = PWMSOUND_ONOFF_PORT -1
.endif

.ifdef PWMSOUND_OSZITRIGGER_PORT
 .equ DDR_PWMSOUND_OSZITRIGGER = PWMSOUND_OSZITRIGGER_PORT-1
.endif

.def dda_deltax_l = r2
.def dda_deltax_h = r3
.def dda_deltay_l = r4
.def dda_deltay_h = r5
.def dda_i_l = r6
.def dda_i_h = r7
.def dda_y_l = r8
; 2010-04-28.def dda_y_h = r7
; .def trigger_zaehler = r10
; .def last_sample = r10
;.if (PWM_ISR_HANDLE_SAMPLE)
; .def current_sample = r9
;.endif

.dseg
PwmSoundKlang:
	.BYTE 1
PwmSoundAktuellerTon:
	.BYTE 1
PwmSoundTimer0Overflows:
	.BYTE 1
PwmSoundPWM_Basis:
	.BYTE 1

.cseg

.include "pwm-sound-samples.inc.asm"
.include "pwm-sound-klaviatur.inc.asm"

PwmSoundInit:
	push  r16
	ldi   r16,127 ; PWM-Frequenz verdoppeln
	sts   PwmSoundPWM_Basis,r16
	ldi   r16,0
	sts   PwmSoundKlang,r16
	sts   PwmSoundAktuellerTon,r16
	; DDR-Port fr Mute-Funktion setzen
.if (PWMSOUND_ONOFF_SCHALTEN)
	sbi   DDR_PWMSOUND_ONOFF, PWMSOUND_ONOFF_PIN
.endif

.if (PWMSOUND_PORTAUSGABE)
	; DDR-Register fr Ausgabe-Pin auf 1 setzen:
	in r16, DDR_PWMSOUND
	ori r16, (1 << PWMSOUND_PIN)
	out DDR_PWMSOUND, r16
.else
	; DDR-Register fr PB3 auf 1 setzen:
	in r16,DDRB
	ori r16,0b00001000
	out DDRB,r16
.endif	
.ifdef PWMSOUND_OSZITRIGGER_PIN
	sbi DDR_PWMSOUND_OSZITRIGGER,PWMSOUND_OSZITRIGGER_PIN
.endif
	pop r16
	ret

PwmSoundTonAus:
	push r16

.if (PWMSOUND_ONOFF_SCHALTEN)
; Verstrker ausschalten:
 .if (PWMSOUND_ONOFF_ON)
 	cbi   PWMSOUND_ONOFF_PORT, PWMSOUND_ONOFF_PIN
 .else
	sbi   PWMSOUND_ONOFF_PORT, PWMSOUND_ONOFF_PIN
 .endif
.endif

	; PWM-Ausgabe und Sampling ausschalten:
	ldi   r16,0b00000000
	out   TCCR2,r16
	in    r16,TIMSK
	andi  r16,~(PWMSOUND_TIMSK)
	out   TIMSK,r16
	ldi   r16,0
	out   OCR2,r16
	out   TCCR0,r16
	pop   r16
	ret

; Schaltet den ber r16 (0..87) indizierten Klavierton ein
PwmSoundTonAn:
	cpi r16,88
	brlo PwmSoundTonAnOK
	ret
PwmSoundTonAnOK:
	push r16
	push r17
	push r30
	push r31
	sts PwmSoundAktuellerTon,r16
	; r16 mit 4 multiplizieren
	ldi r17,4
	mul r16,r17 ; Ergebnis in r0,r1
	; Zum Basisindex der Klaviatur hinzuaddieren:
	ldi r30,LOW(PwmSoundKlaviatur*2)
	ldi r31,HIGH(PwmSoundKlaviatur*2);
	add r30,r0
	adc r31,r1
	; deltax und deltay laden:
	lpm dda_deltax_l,Z+
	lpm dda_deltax_h,Z+
	lpm dda_deltay_l,Z+
	lpm dda_deltay_h,Z

	ldi r16,0
	ldi r17,0
	; mov last_sample,r16
	mov dda_y_l,r16
	movw dda_i_h:dda_i_l,r17:r16
	; mov dda_i_h,r16

	; Timer0 (8 bit) initialisieren
	; r16 ist immer noch 0
	out TCNT0,r16
	sts PwmSoundTimer0Overflows,r16
	; Timer2 (8 bit) initialisieren
	out TCNT2,r16

	inc r16 ; r16 = 1 (CS00=1,CS01=0,CS02=0 => clkIO (no prescaling)
	out TCCR0,r16

	; Timer2 (8 bit mit PWM) initialisieren:
	;  bit		7		6		5		4		3		2		1		0	Seite
	; TCCR2	FOC2	WGM20	COM21	COM20	WGM21	CS22	CS21	CS20	115
	;an:	0		1		1		0		1		0		0		1
	;aus:	0		0		0		0		0		0		0		0
	; FOC (force output compare): 0 <===
	; WGM (wave generation mode): WGM21/WGM20 0/0: Normal; 0/1: PWM, phase correct; 1/0: CTC; 1/1: fast PWM
	; Fast PWM wird ausgewhlt, damit nur das berspringen beim Hochzhlen verhindert werden muss
	; Mode	WGM21(CTC2)	WGM20(PWM2)	Mode of Operation		TOP		Update of OCR2	TOV2 Flag Set
	; 0 	0 			0 			Normal 					0xFF 	Immediate 		MAX
	; 1		0 			1 			PWM, Phase Correct 		0xFF 	TOP 			BOTTOM
	; 2		1			0			CTC 					OCR2	Immediate		MAX
	; 3 	1 			1 			Fast PWM 				0xFF 	TOP 			MAX   <===
	; CS2 (clock select Timer2): 0 0 1
	; Table 46. Clock Select Bit Description
	; CS22 CS21 CS20 Description
	; 0    0    0    No clock source (Timer/Counter stopped).
	; 0    0    1    clkT2S/(No prescaling)                     <===
	; 0    1    0    clkT2S/8 (From prescaler)
	; 0    1    1    clkT2S/32 (From prescaler)
	; 1    0    0    clkT2S/64 (From prescaler)
	; 1    0    1    clkT2S/128 (From prescaler)
	; 1    1    0    clkT2S/256 (From prescaler)
	; 1    1    1    clkT2S/1024 (From prescaler)
	; COM (compare output mode):
	; Table 44. Compare Output Mode, Fast PWM Mode
	; COM21 COM20 	Description
	; 0 	0 		Normal port operation, OC2 disconnected.
	; 0 	1 		Reserved
	; 1 	0 		Clear OC2 on Compare Match, set OC2 at TOP   <===
	; 1 	1 		Set OC2 on Compare Match, clear OC2 at TOP
	ldi r16, (0 << FOC2) | PWMSOUND_WGM2_BITMUSTER | (PWMSOUND_COM2 << COM20) | (PWMSOUND_PRESCALING2 << CS20)
	; frher: ldi r16,0b01101001
	out TCCR2,r16

	;Interrupts fr Timer 0 und 2 Aktivieren:
	in r16,TIMSK
	ori r16,PWMSOUND_TIMSK
	out TIMSK,r16

.if (PWMSOUND_ONOFF_SCHALTEN)
;Verstrker einschalten:
 .if (PWMSOUND_ONOFF_ON)
	sbi   PWMSOUND_ONOFF_PORT, PWMSOUND_ONOFF_PIN
 .else
 	cbi   PWMSOUND_ONOFF_PORT, PWMSOUND_ONOFF_PIN
 .endif
.endif

	pop r31
	pop r30
	pop r17
	pop r16
	ret

PwmSoundNextSample:
	; push r16
	push r17
	; push r18
	; push r19
	; push r20
	push r30 ; ZL
	push r31 ; ZH
	; Oszi Trigger aus:
	;in r16,DDRD
	;ori r16,0b00000010
	;out DDRD,r16
	; cbi PORTD,1
	; cbi PORTB,4

	; ber dda_y_l Sinus-Wert indizieren:
	ldi r30,LOW(PwmSoundSamples*2)
	ldi r31,HIGH(PwmSoundSamples*2)
	lds r16,PwmSoundKlang
	add r31,r16 ; Klang indizieren (Siehe pwm-sound-samples.inc.asm)
	mov r16,dda_y_l
.ifdef PWMSOUND_OSZITRIGGER_PIN
	cpi r16,0
	M_SNE
	sbi PWMSOUND_OSZITRIGGER_PORT,PWMSOUND_OSZITRIGGER_PIN
	M_SEQ
	cbi PWMSOUND_OSZITRIGGER_PORT,PWMSOUND_OSZITRIGGER_PIN
.endif
	ldi r17,0
	add r30,r16
	adc r31,r17
; .if (PWM_ISR_HANDLE_SAMPLE)
;	lpm current_sample,Z
; .else
	lpm r16,Z
	; Den indizierten Wert als PWM-Wert schreiben.
	; Folgendes ist ein Riesenquatsch, da auf Seite 105 steht, dass in den PWM-Modi OCR2 doppelt gepuffert wird,
	; und bei Top oder Bottom synchronisiert wird.
	; aber erst mal checken, ob der Timer2 dadurch nicht berspringen wrde:
	; Das passiert, wenn der aktuelle Zhlerstand kleiner als der aktuelle OCR2-Stand ist und der neue 
	; Wert kleiner als der aktuelle Zhlerstand ist;
	; OCR2 > TCNT2+x > r16
	; bzw r18 > r17+r19 > r16
	; Entsprechend kann der Zhlerstand direkt gesetzt werden, wenn r16 > r18, bzw bersprungen werden, wenn r16 = r18.
	; Weiterhin mssen hier ein paar Takte zustzlich bercksichtigt werden, da der Zhler im folgenden Code 
	; bis "out OCR2,r16" weiter hochzhlt.

	; ldi r19,6 ; 6 Takte a+b+c+d+w
	; ldi r20,3 ; 3 Takte e+f+g
	; in r18,OCR2
	; cp r18,r16
	; breq PwmSoundOCR2Fertig

; PwmSoundNextSWarteschleife:
	; in r17,TCNT2 ; 1 Takt (a)
	; add r17,r19 ; 1 Takt (b)
	; cp r17,r18 ; 1 Takt (c)
	; brsh PwmSoundNextSWeiter; 2 Takte (d)
	; add r17,r20 ; 1 Takt(e)
	; cp r16,r17 ; 1 Takt (f)
	; brlo PwmSoundNextSWarteschleife ; 1 Takt(g)
; PwmSoundNextSWeiter:	
	; das eigentliche schreiben:
	out OCR2,r16 ; 1 Takt (w)
; PwmSoundOCR2Fertig:
; .endif

	; DDA-Agorithmus anwenden:
	ldi r16,1
; 2010-04-28: 	ldi r17,0
	cp dda_deltax_l,dda_deltay_l
	cpc dda_deltax_h,dda_deltay_h
	brpl PwmSounddda_flach
	; Steil oder 45 grad:
PwmSoundNextSFor1:
	add dda_i_l, dda_deltax_l
	adc dda_i_h, dda_deltax_h
	add dda_y_l,r16
; 2010-04-28: 	adc dda_y_h,r17
	cp dda_i_l, dda_deltax_l
	cpc dda_i_h, dda_deltax_h
	brlt PwmSoundNextSFor1
	sub dda_i_l,dda_deltay_l
	sbc dda_i_h,dda_deltay_h
	rjmp PwmSoundNextSEnde
PwmSounddda_flach:
	; flach
	add dda_i_l,dda_deltay_l
	adc dda_i_h,dda_deltay_h
	cp dda_deltay_l,dda_i_l
	cpc dda_deltay_h,dda_i_h
	brge PwmSoundNextSEnde
	sub dda_i_l,dda_deltax_l
	sbc dda_i_h,dda_deltax_h
	add dda_y_l,r16
PwmSoundNextSEnde:
	; Oszi Triggern:
;	cbi PORTD,1
;	sbi PORTB,4

	pop r31
	pop r30
	; pop r20
	; pop r19
	; pop r18
	pop r17
	pop r16
	ret



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Interrupt-Handler:
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

TIM2_COMP: ; Timer2 Compare Handler 
.if (PWMSOUND_PORTAUSGABE)
	sbi PWMSOUND_PORT,PWMSOUND_PIN
.endif
	reti

TIM2_OVF: ; Timer2 Overflow Handler
	push r16
	; Nicht ntig, da keiner der Befehle ein Bit in SREG verndert:
	; in r16,SREG
	; push r16
.if (PWMSOUND_PORTAUSGABE)
	cbi PWMSOUND_PORT,PWMSOUND_PIN
.endif
	lds r16,PwmSoundPWM_Basis
	out TCNT2,r16
;.if (PWM_ISR_HANDLE_SAMPLE)
;	out OCR2,current_sample
;.endif
	; pop r16
	; out SREG,r16
	pop r16
	reti

TIM0_OVF: ; Timer0 Overflow Handler
	sei ; Interrupts wieder ermglichen, damit keine TIM2_OVFs und TIM2_COMPs verschluckt werden
	; Teilt man durch 2, erhlt man eine Sampling-Frequenz von 31250:
	push r16
	in r16,SREG
	push r16
	lds r16,PwmSoundTimer0Overflows
	inc r16
	sts PwmSoundTimer0Overflows,r16
	sbrs r16,0
	rjmp PwmSoundISREnde
	push r17
	push r30 ; ZL
	push r31 ; ZH

	; 31250 mal/sek:
	; ber dda_y_l Sinus-Wert indizieren:
	ldi r30,LOW(PwmSoundSamples*2)
	ldi r31,HIGH(PwmSoundSamples*2)
	lds r16,PwmSoundKlang
	add r31,r16 ; Klang indizieren (Siehe pwm-sound-samples.inc.asm)
	mov r16,dda_y_l
	ldi r17,0
	add r30,r16
	adc r31,r17
	lpm r16,Z
	; Den indizierten Wert als PWM-Wert schreiben.
	; DDA-Agorithmus anwenden:
	ldi r16,1
	cp dda_deltax_l,dda_deltay_l
	cpc dda_deltax_h,dda_deltay_h
	brpl PwmSounddda_flach
	; Steil oder 45 grad:
PwmSoundISRFor1:
	add dda_i_l, dda_deltax_l
	adc dda_i_h, dda_deltax_h
	add dda_y_l,r16
	cp dda_i_l, dda_deltax_l
	cpc dda_i_h, dda_deltax_h
	brlt PwmSoundISRFor1
	sub dda_i_l,dda_deltay_l
	sbc dda_i_h,dda_deltay_h
	rjmp PwmSoundISRNextSEnde
PwmSoundISR_flach:
	; flach
	add dda_i_l,dda_deltay_l
	adc dda_i_h,dda_deltay_h
	cp dda_deltay_l,dda_i_l
	cpc dda_deltay_h,dda_i_h
	brge PwmSoundISRNextSEnde
	sub dda_i_l,dda_deltax_l
	sbc dda_i_h,dda_deltax_h
	add dda_y_l,r16
PwmSoundISRNextSEnde:
	pop r31
	pop r30
	pop r17

PwmSoundISREnde:
	pop r16
	out SREG,r16
	pop r16
	reti


